//********************************************************************
//*** ResourcePool/ResourcePoolOptions.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: ResourcePoolOptions.cc,v 1.26 2009-03-21 18:54:41 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_RESOURCEPOOLOPTIONS_HH
#include "ResourcePoolOptions.hh"
#endif
#include "ResourcePoolEventObserver.hh"

namespace fatalmind {

const ResourcePoolOptions ResourcePoolOptions::defaultOptions;

ResourcePoolOptions::ResourcePoolOptions()
:	_SleepOnFail(1)
,	_Max(5)
,	_MaxTry(2)
,	_MaxExecTry(2)
,	_PreCreate(0)
,	_wait(1)
{
	_SleepOnFail.reserve(5);
}

void ResourcePoolOptions::setMax(const int a_Max) 
{
	if (a_Max < 1) {
		throw IllegalArgumentException(
			"You can not set Max to a value less then 1"
		);
	}	
	_Max = a_Max;
}

void ResourcePoolOptions::setMaxTry(const int a_MaxTry) {
	if (a_MaxTry < 1) {
		throw IllegalArgumentException(
			"You can not set MaxTry to a value less then 1"
		);
	}
	_MaxTry = a_MaxTry;
}

void ResourcePoolOptions::setMaxExecTry(const int a_MaxExecTry) {
	if (a_MaxExecTry < 1) {
		throw IllegalArgumentException(
			"You can not set MaxExecTry to a value less then 1"
		);
	}
	_MaxExecTry = a_MaxExecTry;
}

void ResourcePoolOptions::setPreCreate(const int a_PreCreate) {
	if (a_PreCreate < 0) {
		throw IllegalArgumentException(
			"You can not set PreCreate to a value less the 0"
		);
	}
	_PreCreate = a_PreCreate;
}

void ResourcePoolOptions::setSleepOnFail(const SleepOnFail_t& a_SleepOnFail) {
	if (a_SleepOnFail.size() <= 0) {
		throw IllegalArgumentException(
			"You need to specify at least one SleepOnFail value"
		);
	}
	SleepOnFail_t::const_iterator end = a_SleepOnFail.end();
	for (SleepOnFail_t::const_iterator i = a_SleepOnFail.begin(); i < end; ++i) {
		if (*i < 0) {
			throw IllegalArgumentException(
				"You can not specify a negative value for SleepOnFail"
			);
		}
	}
	_SleepOnFail = a_SleepOnFail;
}

void ResourcePoolOptions::setSleepOnFail(
	  const float s1
	, const float s2
	, const float s3
	, const float s4
) {
	SleepOnFail_t SleepOnFail;
	SleepOnFail.push_back(s1);
	if (s2 != -1) {
		SleepOnFail.push_back(s2);
		if (s3 != -1) {
			SleepOnFail.push_back(s3);
			if (s4 != -1) {
				SleepOnFail.push_back(s4);
			}
		}
	}
	setSleepOnFail(SleepOnFail);
}

POSIX::timespec<SingleThreadedModel<> > ResourcePoolOptions::getSleepOnFail(int tryno) const
{
	int idx = _MaxTry - tryno - 2; // two is the difference of the biggest possible index for a given maxtry
	if (idx < 0 || idx > (static_cast<int>(_SleepOnFail.size()) - 1)) {
		idx = (_SleepOnFail.size() - 1);
	}
	return _SleepOnFail[idx];
}


void ResourcePoolOptions::SleepOnFail(const int tryno) const
{
	POSIX::timespec<SingleThreadedModel<> > t = getSleepOnFail(tryno);
	POSIX::timespec<SingleThreadedModel<> > t2;
	int rv;
	while (((rv = nanosleep(&t, &t2)) == -1) && (errno == OS_EINTR)) {
		t = t2;
	}
}

void ResourcePoolOptions::setWait() {
	_wait = 0;
}

void ResourcePoolOptions::setWait(const float timeout) {
	_wait = timeout;
}

} // namespace fatalmind

