//********************************************************************
//*** ResourcePool/SQL/SQLBatch.t.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLBatch.t.cc,v 1.17 2009-03-19 18:05:40 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

using namespace fatalmind;

void test_SQLBatch_init(RP& p) {
	init(p);
}

void test_SQLBatch_one(RP& p)
{
	RP::Commands::SQLExecute cmd(p, "insert into hirsch (id) values (?)");
	RP::Commands::SQLBatch   b(p);
	cmd.bind(0, 1);
	b.add(cmd);
	cmd.bind(0, 2);
	b.add(cmd);
	p.execute(b);
}

void test_SQLBatch_fail(RP& p)
{
	RP::Commands::SQLExecute cmd(p, "insert into hirsch (id) values (?)");
	RP::Commands::SQLBatch   b(p, true);
	cmd.bind(0, 10);
	b.add(cmd);
	cmd.bind(0, 20);
	b.add(cmd);
	b.add(cmd);
	assertEqual(b.getError(), b.NOT_STARTED);
	try {
		p.execute(b);
		assertNotReach();
	} catch (...) {
	}

	assertTrue(b.getError() >= 0);

	// TODO, test faile in case of syntax error
}

void test_SQLBatch_complex(RP& p)
{
	RP::Commands::SQLExecute clear(p, "delete from hirsch");
	RP::Commands::SQLExecute cmd(p, "insert into hirsch (id, ukn) values (?,?)");
	RP::Commands::SQLSelectRow sel(p, "select ukn from hirsch where id = ?");
	RP::Commands::SQLBatch   b(p, true);

// 0 BATCH POSITION
	b.add(clear);

// 1 BATCH POSITION
	std::string res1 = "res1";
	sel.bind(0, 0);
	sel.bindout(0, res1);
	b.add(sel); // should fail   

// 2 BATCH POSITION
	cmd.bind(0, 10);
	cmd.bind<std::string>(1, "hirsch 1");
	b.add(cmd); // intrnal pos 0

// 3 BATCH POSITION
	cmd.bind(0, 20);
	cmd.bind<std::string>(1, "hirsch 20");
	b.add(cmd); // intrnal pos 0

// 4 BATCH POSITION
	RP::Commands::SQLSelectRow sel2(p, "select ukn from hirsch where id = ?");
	std::string res2 ="res2";
	sel2.bind(0, 10);
	sel2.bindout(0, res2);
	b.add(sel2);

// 5 BATCH POSITION
	cmd.bind(0, 100);
	cmd.bind<std::string>(1, "hirsch 100");
	b.add(cmd); // intrnal pos 0

// 6 BATCH POSITION
	cmd.bind(0, 101);
	cmd.bind<std::string>(1, "hirsch 101");
	b.add(cmd); // intrnal pos 1

// 7 BATCH POSITION
	cmd.bind(0, 102);
	cmd.bind<std::string>(1, "hirsch 102");
	b.add(cmd); // intrnal pos 2

// 8 BATCH POSITION
	cmd.bind(0, 100);
	cmd.bind<std::string>(1, "hirsch 100");
	b.add(cmd); // should fail // intrnal pos 3

// 9 BATCH POSITION
	cmd.bind(0, 103);
	cmd.bind<std::string>(1, "hirsch 103");
	b.add(cmd); // intrnal pos 4

// 10 BATCH POSITION
	std::string res3 = "res3";
	sel.bind(0, 105);
	sel.bindout(0, res3); // should fail
	b.add(sel);

// 11 BATCH POSITION
	std::string res4 = "res4";
	sel.bind(0, 100);
	sel.bindout(0, res4); 
	b.add(sel);

	while (b.getExecPosition() != b.end()) {
		try {
			p.execute(b);
		} catch (SQL::SQLException e) {
			int offset =  b.getError();
			const RP::Commands::Base& failed = **(b.begin() + offset);
		}
	}

	assertEqual(res2, "hirsch 1"); // BATCH POS 4 worked
	assertEqual(res4, "hirsch 100"); // BATCH POS 11 worked
	assertEqual(res1, "res1"); // BATCH POS 1 failed
	assertEqual(res3, "res3"); // BATCH POS 10 failed
}

void test_SQLBatch_nested(RP& p) {
	RP::Commands::SQLExecute clear(p, "delete from hirsch");
	RP::Commands::SQLExecute cmd(p, "insert into hirsch (id, ukn) values (?,?)");
	RP::Commands::SQLSelectRow sel(p, "select ukn from hirsch where id = ?");
	RP::Commands::SQLBatch   b(p, true);
	RP::Commands::SQLBatch   b2(p, true);

	b.add(clear);
	
	cmd.bind(0, 0);
	cmd.bind<std::string>(1, "hirsch");
	b2.add(cmd);

	cmd.bind(0, 0);
	cmd.bind<std::string>(1, "hirsch");
	b2.add(cmd); // will fail

	cmd.bind(0, 1);
	cmd.bind<std::string>(1, "hirsch");
	b2.add(cmd);

	cmd.bind(0, 1);
	cmd.bind<std::string>(1, "hirsch");
	b2.add(cmd);

	b.add(b2); // pos 1 in b

	while (b.getExecPosition() != b.end()) {
		try {
			p.execute(b);
		} catch (SQL::SQLException e) {
			int offset =  b.getError();
			const RP::Commands::Base& failed = **(b.begin() + offset);
		}
	}
}

template<class B>
void test_PERF_SQLBatch_ok(RP& p) {
	RP::Commands::SQLExecute clear(p, "delete from hirsch");
	RP::Commands::SQLExecute cmd(p, "insert into hirsch (id, ukn) values (?,?)");
	B   b(p, true);

	b.add(clear);
	ostringstream oss;
	std::string empty;
	for (int i = 0; i < 5000; ++i) {
		oss.str(empty);
		oss << "record " << i;
		cmd.bind(0, i);
		cmd.bind(1, oss.str());
		b.add(cmd);
	}

	p.execute(b);
}

void test_PERF_SQLBatch_failoften(RP& p) {
	RP::Commands::SQLExecute clear(p, "delete from hirsch");
	RP::Commands::SQLExecute cmd(p, "insert into hirsch (id, ukn) values (?,?)");
	RP::Commands::SQLBatch  b(p, true);

	b.add(clear);
	int cnt = 5000;
	for (int i = 0; i < cnt; ++i) {
		ostringstream oss;
		oss << "record " << i;
		cmd.bind(0, i/2);
		cmd.bind(1, oss.str());
		b.add(cmd);
	}

	int i = 2;
	int exceptions = 0;
	assertEqual(std::distance(b.begin(), b.end()), cnt+1);
	assertEqual(std::distance(b.begin(), b.getExecPosition()), 0);
	while (b.getExecPosition() != b.end()) {
		try {
			p.execute(b);
		} catch (SQL::SQLException e) {
			exceptions++;
			assertEqual(i, b.getError());
			i+=2;
		}
	}
	assertEqual(exceptions, cnt/2);
}
