//********************************************************************
//*** ResourcePool/mysql/SQLExecute.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLExecute.cc,v 1.35 2009-03-17 19:13:54 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */


#include "SQLExecute.hh"

#include "mysql_bind.hh"
//#include "ResourcePool/ResourcePool.hh"

#ifndef INCLUDED_EXCEPTIONS_HH
#include "Exceptions/Exception.hh"
#endif

#ifndef RESOURCEPOOL_MYSQL_SQLTYPE_HH
#include "SQLType.hh"
#endif

#ifndef RESOURCEPOOL_MYSQL_MYSQLEXCEPTION_HH
#include "MysqlException.hh"
#endif

#include <mysql/mysql.h>

namespace fatalmind {
namespace mysql {

SQLExecute::SQLExecute(const SQL::SQLStatement& SQL)
: _SQL(SQL)
{
}

SQLExecute::SQLExecute(const ResourcePool<ResourceType>& p, const SQL::SQLStatement& SQL)
: _SQL(SQL)
{
}

SQLExecute::~SQLExecute()
{
}

Clone* SQLExecute::DoClone() const 
{
	return new SQLExecute(*this);
}

class EscalateWarnings: public SQLSelect, public SQL::SQLFetcher {
	public:
		EscalateWarnings() 
		: SQLSelect("SHOW WARNINGS", *this)
		, level()
		, code(0)
		, message()
		, cnt(0)
		, allmessages()
		{
			bindout(0, level);
			bindout(1, code);
			bindout(2, message);
		}

		virtual bool process(int row) {
			switch(code) {
			case 1263:
				++cnt;
				allmessages.append("[");
				allmessages.append(message);
				allmessages.append("]");
				break;
			}
			return true;
		}

		int getCriticalWarnings() const {
			return cnt;
		}

		std::string getMessages() const {
			return allmessages;
		}

		EscalateWarnings* DoClone() const {
			return new EscalateWarnings(*this);
		}

	private:
		std::string level;
		int code;
		std::string message;

		int cnt;
		std::string allmessages;
};

void SQLExecute::execute(mysqlResource& rdbh)
{
	MYSQL& dbh = rdbh.get_plain_resource();
	mysql_bind boundsql(dbh, _SQL, _args);
	std::string mysql_sql = boundsql.getSQL();
	if (int rc = mysql_real_query(&dbh
		, mysql_sql.c_str()
		, mysql_sql.length())
	) {
		gc_ptr<EventReader> er = rdbh.getReader(rdbh.Event.EXECUTE);
		(*er) << mysql_errno(&dbh) << ";" << (*this);
		MysqlException::throwit(rc, mysql_error(&dbh), mysql_errno(&dbh), _SQL);
	}

// mysql_warning_count was introduced with 4.1
#ifdef HAVE_MYSQL_WARNINGCOUNT
	if (mysql_warning_count(&dbh) > 0) {
		// a warning happend
		EscalateWarnings warnings;
		warnings.execute(rdbh);
		if (warnings.getCriticalWarnings() > 0) {
			gc_ptr<EventReader> er = rdbh.getReader(rdbh.Event.EXECUTE);
			(*er) << "CRITICAL WARNING;" << (*this);
			throw MysqlException(warnings.getMessages(), 0, _SQL);
		}
	}
#endif
	(*rdbh.getReader(rdbh.Event.EXECUTE)) << ';' << (*this); // no ERROR
}

void SQLExecute::outputoperator(std::ostream& s) const {
	s << "mysql::SQLExecute[" << _SQL << '|';
	bindoutputoperator(s);
	s << ']';
}

} // namespace mysql

template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<char>
	(const unsigned int pos, const char& val, bool nul);
template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<signed char>
	(const unsigned int pos, const signed char& val, bool nul);
template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<unsigned char>
	(const unsigned int pos, const unsigned char& val, bool nul);

template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<short int>
	(const unsigned int pos, const short int& val, bool nul);
template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<int>	
	(const unsigned int pos, const int& val, bool nul);
template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<long int>
	(const unsigned int pos, const long int& val, bool nul);

template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<unsigned short int>
	(const unsigned int pos, unsigned const short int& val, bool nul);
template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<unsigned int>
	(const unsigned int pos, unsigned const int& val, bool nul);
template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<unsigned long int>
	(const unsigned int pos, unsigned const long int& val, bool nul);

// wchar_t, bool, not supported..

template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<float>
	(const unsigned int pos, const float& val, bool nul);
template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<double>
	(const unsigned int pos, const double& val, bool nul);
template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<long double>
	(const unsigned int pos, const long double& val, bool nul);

template void SQL::Bind<mysql::SQLType, mysql::SQLTypeImpl>::bind<std::string>
	(const unsigned int pos, const std::string& val, bool nul);



} // namespace fatalmind
